# -*- coding: utf-8 -*-
##
## This file is part of the SixPay Indico EPayment Plugin.
## Copyright (C) 2017 - 2018 Max Fischer
##
## This is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License as
## published by the Free Software Foundation; either version 3 of the
## License, or (at your option) any later version.
##
## This software is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
## General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with SixPay Indico EPayment Plugin;if not, see <http://www.gnu.org/licenses/>.



"""
++++++++++++++++++++++++++++++++++++++++++++++++++
``indico_sixpay`` - SIX EPayment Plugin for Indico
++++++++++++++++++++++++++++++++++++++++++++++++++

.. image:: https://readthedocs.org/projects/indico_sixpay/badge/?version=latest
    :target: http://indico-sixpay.readthedocs.io/en/latest/?badge=latest
    :alt: Documentation

.. image:: https://img.shields.io/pypi/v/indico_sixpay.svg
    :alt: Available on PyPI
    :target: https://pypi.python.org/pypi/indico_sixpay/

.. image:: https://img.shields.io/github/license/maxfischer2781/indico_sixpay.svg
    :alt: License
    :target: https://github.com/maxfischer2781/indico_sixpay/blob/master/LICENSE

.. image:: https://img.shields.io/github/commits-since/maxfischer2781/indico_sixpay/v2.0.0.svg
    :alt: Repository
    :target: https://github.com/maxfischer2781/indico_sixpay/tree/master

Plugin for the Indico event management system to use EPayment via SIX Payment services.

Quick Guide
-----------

To enable the plugin, it must be installed for the python version running ``indico``.

.. code:: bash

    python -m pip install indico_sixpay

Once installed, it can be enabled in the administrator and event settings.
Configuration uses the same options for global defaults and event specific overrides.

Disclaimer
----------

This plugin is in no way endorsed, supported or provided by SIX, Indico or any other service, provider or entity.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""
__title__ = 'indico_sixpay'
__summary__ = 'Indico EPayment Plugin for SixPay services'
__url__ = 'https://github.com/maxfischer2781/indico_sixpay'

__version__ = '2.0.1'
__author__ = 'Max Fischer'
__email__ = 'maxfischer2781@gmail.com'
__copyright__ = '2017 - 2018 %s' % __author__



# -*- coding: utf-8 -*-
##
## This file is part of the SixPay Indico EPayment Plugin.
## Copyright (C) 2017 - 2018 Max Fischer
##
## This is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License as
## published by the Free Software Foundation; either version 3 of the
## License, or (at your option) any later version.
##
## This software is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
## General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with SixPay Indico EPayment Plugin;if not, see <http://www.gnu.org/licenses/>.
"""
Definition of callbacks exposed by the Indico server
"""
from __future__ import unicode_literals

from indico.core.plugins import IndicoPluginBlueprint

from .request_handlers import SixPayResponseHandler, UserCancelHandler, UserFailureHandler, UserSuccessHandler


#: url mount points exposing callbacks
blueprint = IndicoPluginBlueprint(
    'payment_sixpay', __name__,
    url_prefix='/event/<confId>/registrations/<int:reg_form_id>/payment/response/sixpay'
)

blueprint.add_url_rule('/failure', 'failure', UserCancelHandler, methods=('GET', 'POST'))
blueprint.add_url_rule('/cancel', 'cancel', UserFailureHandler, methods=('GET', 'POST'))
blueprint.add_url_rule('/success', 'success', UserSuccessHandler, methods=('GET', 'POST'))
# Used by SixPay to send an asynchronous notification for the transaction
blueprint.add_url_rule('/ipn', 'notify', SixPayResponseHandler, methods=('Get', 'POST'))



import os
from setuptools import setup, find_packages

repo_base_dir = os.path.abspath(os.path.dirname(__file__))
# pull in the packages metadata
package_about = {}
with open(os.path.join(repo_base_dir, "indico_sixpay", "__about__.py")) as about_file:
    exec(about_file.read(), package_about)

setup(
    name=package_about['__title__'],
    version=package_about['__version__'],
    description=package_about['__summary__'],
    long_description=package_about['__doc__'].strip(),
    author=package_about['__author__'],
    author_email=package_about['__email__'],
    url=package_about['__url__'],
    entry_points={
        'indico.plugins': {
            'payment_sixpay = indico_sixpay.plugin:SixpayPaymentPlugin'
        }
    },
    packages=find_packages(),
    package_data={'indico_sixpay': ['templates/*.html']},
    install_requires=['requests', 'indico>=2.0', 'iso4217'],
    license='GPLv3+',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Plugins',
        'Environment :: Web Environment',
        'Intended Audience :: Science/Research',
        'Intended Audience :: System Administrators',
        'Intended Audience :: Other Audience',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Topic :: Communications :: Conferencing',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Application',
    ],
    zip_safe=False,
    keywords='indico epayment six sixpay plugin',
)



# -*- coding: utf-8 -*-
##
## This file is part of the SixPay Indico EPayment Plugin.
## Copyright (C) 2017 - 2018 Max Fischer
##
## This is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License as
## published by the Free Software Foundation; either version 3 of the
## License, or (at your option) any later version.
##
## This software is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
## General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with SixPay Indico EPayment Plugin;if not, see <http://www.gnu.org/licenses/>.
"""
Callbacks for asynchronous replies by the SixPay service and to redirect the user
"""
from __future__ import unicode_literals
import urlparse
from xml.dom.minidom import parseString

import requests
from flask import flash, redirect, request
from flask_pluginengine import current_plugin
from werkzeug.exceptions import BadRequest

from indico.modules.events.payment.models.transactions import TransactionAction
from indico.modules.events.payment.notifications import notify_amount_inconsistency
from indico.modules.events.payment.util import register_transaction
from indico.modules.events.registration.models.registrations import Registration
from indico.web.flask.util import url_for
from indico.web.rh import RH

from .utility import gettext, to_large_currency, to_small_currency

# RH from indico.web.rh
# - the logic to execute when SixPay/Users are redirected *after* a transaction
# - see blueprint.py for how RH's are mounted!


class BaseRequestHandler(RH):
    """
    Request Handler for asynchronous callbacks from SixPay

    These handlers are used either by

    - the user, when he is redirected from SixPay back to Indico
    - SixPay, when it sends back the result of a transaction
    """
    CSRF_ENABLED = False

    def _process_args(self):
        self.token = request.args['token']
        self.registration = Registration.find_first(uuid=self.token)
        if not self.registration:
            raise BadRequest


class TransactionFailure(Exception):
    """
    A Transaction with SixPay failed

    :param step: name of the step at which the transaction failed
    :type step: basestring
    :param details: verbose description of what went wrong
    :type step: basestring
    """
    def __init__(self, step, details=None):
        self.step = step
        self.details = details


class SixPayResponseHandler(BaseRequestHandler):
    """Handler for notification from SixPay service"""
    def __init__(self):
        super(SixPayResponseHandler, self).__init__()
        # registration context is not initialised before `self._process_args`...
        self.sixpay_url = None  # type: str

    def _process_args(self):
        super(SixPayResponseHandler, self)._process_args()
        # prefer event supplied sixpay url over global sixpay url
        self.sixpay_url = current_plugin.event_settings.get(self.registration.registration_form.event, 'url') \
            or current_plugin.settings.get('url')

    def _process(self):
        """process the reply from SixPay about the transaction"""
        try:
            self._process_confirmation()
        except TransactionFailure as err:
            current_plugin.logger.warning("SixPay transaction failed during %s: %s" % (err.step, err.details))

    def _process_confirmation(self):
        """Process the confirmation response inside indico"""
        # DATA: '<IDP
        #           MSGTYPE="PayConfirm" TOKEN="(unused)" VTVERIFY="(obsolete)" KEYID="1-0"
        #           ID="9SUO5zbOGY6OUA45b222bhvrpW2A"
        #           ACCOUNTID="401860-17795278"
        #           PROVIDERID="90"
        #           PROVIDERNAME="Saferpay Test Card"
        #           PAYMENTMETHOD="6"
        #           ORDERID="c281r4"
        #           AMOUNT="100"
        #           CURRENCY="EUR"
        #           IP="141.3.200.120"
        #           IPCOUNTRY="DE"
        #           CCCOUNTRY="US"
        #           MPI_LIABILITYSHIFT="yes"
        #           MPI_TX_CAVV="jAABBIIFmAAAAAAAAAAAAAAAAAA="
        #           MPI_XID="VVE3DQlhXR8PBD5JPzYGWW5FNgI="
        #           ECI="1"
        #           CAVV="jAABBIIFmAAAAAAAAAAAAAAAAAA="
        #           XID="VVE3DQlhXR8PBD5JPzYGWW5FNgI=" />'
        transaction_xml = request.args['DATA']
        transaction_signature = request.args['SIGNATURE']
        transaction_data = self._parse_transaction_xml(transaction_xml)
        # verify the signature of SixPay for the transaction
        # if this matches, the user completed the transaction as requested by Indico
        try:
            self._verify_signature(transaction_xml, transaction_signature, transaction_data['ID'])
            if self._is_duplicate_transaction(transaction_data=transaction_data):
                # we have already handled the transaction
                return True
            if self._confirm_transaction(transaction_data):
                self._verify_amount(transaction_data)
                self._register_transaction(transaction_data)
        except TransactionFailure as err:
            current_plugin.logger.warning("SixPay transaction failed during %s: %s" % (err.step, err.details))
            raise
        return True

    def _perform_request(self, task, endpoint, **data):
        """
        Helper for performing a request against SixPay

        :param task: description of the request, used for error handling
        :type task: basestring
        :param endpoint: the URL endpoint *relative* to the SixPay base URL
        :type endpoint: basestring
        :param **data: data passed during the request

        This will automatically raise any HTTP errors encountered during the request.
        If the request itself fails, a :py:exc:`~.TransactionFailure` is raised for ``task``.
        """
        request_url = urlparse.urljoin(self.sixpay_url, endpoint)
        response = requests.post(request_url, data)
        response.raise_for_status()
        if response.text.startswith('ERROR'):
            raise TransactionFailure(step=task, details=response.text)
        return response.text

    @staticmethod
    def _parse_transaction_xml(transaction_xml):
        """Parse the ``transaction_xml`` to a mapping"""
        mdom = parseString(transaction_xml)
        attributes = mdom.documentElement.attributes
        idp_data = {
            attributes.item(idx).name: attributes.item(idx).value
            for idx in range(attributes.length)
        }
        return idp_data

    def _verify_signature(self, transaction_xml, transaction_signature, transaction_id):
        """Verify the transaction data and signature with SixPay"""
        verification_response = self._perform_request(
            'verification', 'VerifyPayConfirm.asp',
            DATA=transaction_xml, SIGNATURE=transaction_signature
        )
        if verification_response.startswith('OK'):
            # text = 'OK:ID=56a77rg243asfhmkq3r&TOKEN=%3e235462FA23C4FE4AF65'
            content = verification_response.split(':', 1)[1]
            confirmation = dict(key_value.split('=') for key_value in content.split('&'))
            if not confirmation['ID'] == transaction_id:
                raise TransactionFailure(step='verification', details='mismatched transaction ID')
            return True
        raise RuntimeError("Expected reply 'OK:ID=...&TOKEN=...', got %r" % verification_response.text)

    def _is_duplicate_transaction(self, transaction_data):
        """Check if this transaction has already been recorded"""
        prev_transaction = self.registration.transaction
        if not prev_transaction or prev_transaction.provider != 'sixpay':
            return False
        return all(
            prev_transaction.data.get(key) == transaction_data.get(key)
            for key in ('ORDERID', 'CURRENCY', 'AMOUNT', 'ACCOUNTID')
        )

    def _verify_amount(self, transaction_data):
        """Verify the amount and currency of the payment; sends an email but still registers incorrect payments"""
        expected_amount = float(self.registration.price)
        expected_currency = self.registration.currency
        amount = float(transaction_data['AMOUNT'])
        currency = transaction_data['CURRENCY']
        if to_small_currency(expected_amount, expected_currency) == amount and expected_currency == currency:
            return True
        current_plugin.logger.warning(
            "Payment doesn't match events fee: %s %s != %s %s",
            amount, currency, to_small_currency(expected_amount, expected_currency), expected_currency
        )
        notify_amount_inconsistency(self.registration, to_large_currency(amount, currency), currency)
        return False

    def _confirm_transaction(self, transaction_data):
        """Confirm to SixPay that the transaction is accepted"""
        completion_data = {'ACCOUNTID': transaction_data['ACCOUNTID'], 'ID': transaction_data['ID']}
        if 'test.saferpay.com' in self.sixpay_url:
            # password: see "Saferpay Payment Page" specification, v5.1, section 4.6
            completion_data['spPassword'] = '8e7Yn5yk'
        completion_response = self._perform_request('confirmation', 'PayCompleteV2.asp', **completion_data)
        assert completion_response.startswith('OK')
        return True

    def _register_transaction(self, transaction_data):
        """Register the transaction persistently within Indico"""
        register_transaction(
            registration=self.registration,
            # SixPay uses SMALLEST currency, Indico expects LARGEST currency
            amount=to_large_currency(float(transaction_data['AMOUNT']), transaction_data['CURRENCY']),
            currency=transaction_data['CURRENCY'],
            action=TransactionAction.complete,
            provider='sixpay',
            data=transaction_data,
        )


class UserCancelHandler(BaseRequestHandler):
    """User Message on cancelled payment"""
    def _process(self):
        flash(gettext('You cancelled the payment.'), 'info')
        return redirect(url_for('event_registration.display_regform', self.registration.locator.registrant))


class UserFailureHandler(BaseRequestHandler):
    """User Message on failed payment"""
    def _process(self):
        flash(gettext('Your payment has failed.'), 'info')
        return redirect(url_for('event_registration.display_regform', self.registration.locator.registrant))


class UserSuccessHandler(SixPayResponseHandler):
    """User Message on successful payment"""
    def _process(self):
        try:
            self._process_confirmation()
        except TransactionFailure as err:
            current_plugin.logger.warning("SixPay transaction failed during %s: %s" % (err.step, err.details))
            flash(gettext('Your payment could not be confirmed. Please contact an organizer.'), 'info')
        else:
            flash(gettext('Your payment has been confirmed.'), 'success')
        return redirect(url_for('event_registration.display_regform', self.registration.locator.registrant))



# -*- coding: utf-8 -*-
##
## This file is part of the SixPay Indico EPayment Plugin.
## Copyright (C) 2017 - 2018 Max Fischer
##
## This is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License as
## published by the Free Software Foundation; either version 3 of the
## License, or (at your option) any later version.
##
## This software is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
## General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with SixPay Indico EPayment Plugin;if not, see <http://www.gnu.org/licenses/>.
"""
Core of the SixPay plugin

The entry point for indico is the :py:class:`~.SixpayPaymentPlugin`.
It handles configuration via the settings forms, initiates payments
and provides callbacks for finished payments via its blueprint.
"""
from __future__ import unicode_literals, absolute_import
import urlparse

import requests
from werkzeug.exceptions import NotImplemented as HTTPNotImplemented, InternalServerError as HTTPInternalServerError

from wtforms.fields.core import StringField
from wtforms.fields.html5 import URLField
from wtforms.validators import DataRequired, Optional, Regexp, Length, Email, ValidationError

from indico.core.plugins import IndicoPlugin, url_for_plugin
from indico.modules.events.payment import \
    PaymentEventSettingsFormBase, PaymentPluginMixin, PaymentPluginSettingsFormBase

from .utility import gettext, to_small_currency
# blueprint mounts the request handlers onto URLs
from .blueprint import blueprint


# Dear Future Maintainer,
#
# while an improvement over Indico 1.2, the Indico 2.0 plugin/core
# facilities are rather lacking in documentation. I have added
# some notes for each **base**type on how they are integrated with
# the rest of Indico. Be aware that this is reconstructed from
# implementations, not from official API docs.
#
# Regards,
# Past Maintainer


# PaymentPluginSettingsFormBase from indico.modules.events.payment
# - A codified Form for users to fill in. The *class attributes* define
#   which fields exist, their shape, description, etc.
# - Each field is a type from wtforms.fields.core.Field. You probably want to have:
#   - label: Name of the field, an internationalised identifier
#   - validators: Input validation, see wtforms.validators
#   - description: help text of the field, an internationalised text

class FormatField(object):
    """
    Validator for format fields, i.e. strings with ``{key}`` placeholders

    :param max_length: optional maximum length, checked on a test formatting
    :type max_length: int
    :param field_map: keyword arguments to use for test formatting

    On validation, a test mapping is applied to the field. This ensures the
    field has a valid ``str.format`` format, and does not use illegal keys
    (as determined by ``default_field_map`` and ``field_map``).
    The ``max_length`` is validated against the test-formatted field, which
    is an estimate for an average sized input.
    """
    #: default placeholders to test length after formatting
    default_field_map = {
        'user_id': 1234,
        'user_name': 'Jane Whiteacre',
        'event_id': 123,
        'event_title': 'Placeholder: The Event',
        'eventuser_id': 'e123u1234',
        'registration_title': 'EarlyBird Registration'
    }

    def __init__(self, max_length=float('inf'), field_map=None):
        self.max_length = max_length
        self.field_map = self.default_field_map.copy()
        if field_map is not None:
            self.field_map.update(field_map)

    def __call__(self, form, field):
        if not field.data:
            return True
        try:
            test_format = field.data.format(**self.field_map)
        except KeyError as err:
            raise ValidationError('Invalid format string key: {}'.format(err))
        except ValueError as err:
            raise ValidationError('Malformed format string: {}'.format(err))
        if len(test_format) > self.max_length:
            raise ValidationError(
                'Too long format string: shortest replacement with {0}, expected {1}'.format(
                    len(test_format), self.max_length
                )
            )
        else:
            return True


class PluginSettingsForm(PaymentPluginSettingsFormBase):
    """Configuration form for the Plugin across all events"""
    url = URLField(
        gettext('SixPay Saferpay URL'),
        [DataRequired()],
        description=gettext('URL to contact the Six Payment Service'),
    )
    account_id = StringField(
        label='Account ID',
        # can be set EITHER or BOTH globally and per event
        validators=[Optional(), Regexp(r'[0-9-]{0,15}', message='Field must contain up to 15 digits or "-".')],
        description=gettext('Default ID of your Saferpay account, such as "401860-17795278".')
    )
    order_description = StringField(
        label=gettext('Order Description'),
        validators=[DataRequired(), FormatField(max_length=80)],
        description=gettext('The description of each order in a human readable way.'
                            'This description is presented to the registrant during the transaction with SixPay.')
    )
    order_identifier = StringField(
        label=gettext('Order Identifier'),
        validators=[DataRequired(), FormatField(max_length=80)],
        description=gettext('The identifier of each order for further processing.')
    )
    notification_mail = StringField(
        label=gettext('Notification Email'),
        validators=[Optional(), Email(), Length(0, 50)],
        description=gettext('Mail address to receive notifications of transactions.'
                            'This is independent of Indico\'s own payment notifications.')
    )


class EventSettingsForm(PaymentEventSettingsFormBase):
    """Configuration form for the Plugin for a specific event"""
    # every setting may be overwritten for each event
    url = PluginSettingsForm.url
    account_id = PluginSettingsForm.account_id
    order_description = PluginSettingsForm.order_description
    order_identifier = PluginSettingsForm.order_identifier
    notification_mail = PluginSettingsForm.notification_mail


# PaymentPluginMixin, IndicoPlugin
# This is basically a registry of setting fields, logos and other rendering stuff
# All the business logic is in :py:func:`adjust_payment_form_data`
class SixpayPaymentPlugin(PaymentPluginMixin, IndicoPlugin):
    """
    SixPay Saferpay

    Provides an EPayment method using the SixPay Saferpay API.
    """
    configurable = True
    #: form for default configuration across events
    settings_form = PluginSettingsForm
    #: form for configuration for specific events
    event_settings_form = EventSettingsForm
    #: global default settings - should be a reasonable default
    default_settings = {
        'method_name': 'SixPay',
        'url': 'https://www.saferpay.com/hosting',
        'account_id': None,
        'order_description': '{event_title}, {registration_title}, {user_name}',
        'order_identifier': '{eventuser_id}',
        'notification_mail': None
    }
    #: per event default settings - use the global settings
    default_event_settings = {
        'enabled': False,
        'method_name': None,
        'url': None,
        'account_id': None,
        'order_description': None,
        'order_identifier': None,
        'notification_mail': None,
    }

    def get_blueprints(self):
        """Blueprint for URL endpoints with callbacks"""
        return blueprint

    # Dear Future Maintainer,
    # - business logic is here!
    # - see PaymentPluginMixin.render_payment_form for what `data` provides
    # - What happens here
    #   - We already send all payment details to SixPay to get a *signed* request url for this transaction
    #   - We have added `success`, `cancel` and `failure` for *sixpay* to redirect the user back to us AFTER his request
    #   - We have added `notify` for *sixpay* to inform us asynchronously about the result
    #   - We put the transaction URL we got into `data` for the *user* to perform his request securely
    #   - Return uses `indico_sixpay/templates/event_payment_form.html`, presenting a trigger button to the user
    def adjust_payment_form_data(self, data):
        """Prepare the payment form shown to registrants"""
        registration = data['registration']
        # indico does not seem to provide stacking of settings
        # we merge event on top of global settings, but remove placeholder defaults
        event_settings, global_settings = data['event_settings'], data['settings']
        plugin_settings = {
            key: event_settings[key] if event_settings.get(key) is not None else global_settings[key]
            for key in
            (set(event_settings) | set(global_settings))
        }
        # parameters of the transaction - amount, currency, ...
        transaction = self._get_transaction_parameters(data)
        # callbacks of the transaction - where to announce success, failure, ...
        # where to redirect the user
        transaction['SUCCESSLINK'] = url_for_plugin('payment_sixpay.success', registration.locator.uuid, _external=True)
        transaction['BACKLINK'] = url_for_plugin('payment_sixpay.cancel', registration.locator.uuid, _external=True)
        transaction['FAILLINK'] = url_for_plugin('payment_sixpay.failure', registration.locator.uuid, _external=True)
        # where to asynchronously call back from SixPay
        transaction['NOTIFYURL'] = url_for_plugin('payment_sixpay.notify', registration.locator.uuid, _external=True)
        data['payment_url'] = self._get_payment_url(sixpay_url=plugin_settings.get('url'), transaction_data=transaction)
        return data

    @staticmethod
    def get_field_format_map(registration):
        """Generates dict which provides registration information for format fields"""
        return {
            'user_id': registration.user_id,
            'user_name': registration.full_name,
            'user_firstname': registration.first_name,
            'user_lastname': registration.last_name,
            'event_id': registration.event_id,
            'event_title': registration.event.title,
            'eventuser_id': 'e{0}u{1}'.format(registration.event_id, registration.user_id),
            'registration_title': registration.registration_form.title
        }

    def _get_transaction_parameters(self, payment_data):
        """Parameters for formulating a transaction request *without* any business logic hooks"""
        plugin_settings = payment_data['event_settings']
        format_map = self.get_field_format_map(payment_data['registration'])
        for format_field in 'order_description', 'order_identifier':
            try:
                if not plugin_settings.has_key(format_field):
                    raise KeyError
                payment_data[format_field] = plugin_settings.get(format_field).format(**format_map)
            except ValueError:
                message = "Invalid format field placeholder for {0}, please contact the event organisers!"
                raise HTTPNotImplemented((
                        gettext(message) + '\n\n[' + message + ']'
                    ).format(self.name)
                 )
            except KeyError:
                message = "Unknown format field placeholder '{0}' for {1}, please contact the event organisers!"
                raise HTTPNotImplemented((
                        gettext(message) + '\n\n[' + message + ']'
                    ).format(format_field, self.name)
                )
        # see the SixPay Manual on what these things mean
        transaction_parameters = {
            'ACCOUNTID': str(plugin_settings.get('account_id')),
            # indico handles price as largest currency, but six expects smallest
            # e.g. EUR: indico uses 100.2 Euro, but six expects 10020 Cent
            'AMOUNT': '{:.0f}'.format(to_small_currency(payment_data['amount'], payment_data['currency'])),
            'CURRENCY': payment_data['currency'],
            'DESCRIPTION': payment_data['order_description'][:50],
            'ORDERID': payment_data['order_identifier'][:80],
            'SHOWLANGUAGES': 'yes',
        }
        if plugin_settings.get('notification_mail'):
            transaction_parameters['NOTIFYADDRESS'] = plugin_settings.get('notification_mail')
        return transaction_parameters

    def _get_payment_url(self, sixpay_url, transaction_data):
        """Send transaction data to SixPay to get a signed URL for the user request"""
        endpoint = urlparse.urljoin(sixpay_url, 'CreatePayInit.asp')
        url_request = requests.post(endpoint, data=transaction_data)
        # raise any HTTP errors
        url_request.raise_for_status()
        if url_request.text.startswith('ERROR'):
            raise HTTPInternalServerError('Failed request to SixPay service: %s' % url_request.text)
        return url_request.text



# -*- coding: utf-8 -*-
##
## This file is part of the SixPay Indico EPayment Plugin.
## Copyright (C) 2017 - 2018 Max Fischer
##
## This is free software; you can redistribute it and/or
## modify it under the terms of the GNU General Public License as
## published by the Free Software Foundation; either version 3 of the
## License, or (at your option) any later version.
##
## This software is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
## General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with SixPay Indico EPayment Plugin;if not, see <http://www.gnu.org/licenses/>.
from __future__ import unicode_literals, division

import iso4217
from werkzeug.exceptions import NotImplemented as HTTPNotImplemented

from indico.util.i18n import make_bound_gettext

#: internationalisation/localisation of strings
gettext = make_bound_gettext('payment_sixpay')


#: currencies for which the major to minor currency ratio is not a multiple of 10
NON_DECIMAL_CURRENCY = {'MRU', 'MGA'}


def validate_currency(iso_code):
    """
    Check whether the currency can be properly handled by this plugin

    :param iso_code: an ISO4217 currency code, e.g. ``"EUR"``
    :type iso_code: basestring
    :raises: :py:exc:`~.HTTPNotImplemented` if the currency is not valid
    """
    if iso_code in NON_DECIMAL_CURRENCY:
        raise HTTPNotImplemented(
            gettext("Unsupported currency '{0}' for SixPay. Please contact the organisers").format(iso_code)
        )
    try:
        iso4217.Currency(iso_code)
    except ValueError:
        raise HTTPNotImplemented(
            gettext("Unknown currency '{0}' for SixPay. Please contact the organisers").format(iso_code)
        )


def to_small_currency(large_currency_amount, iso_code):
    """
    Convert an amount from large currency to small currency, e.g. 2.3 Euro to 230 Eurocent

    :param large_currency_amount: the amount in large currency, e.g. ``2.3``
    :param iso_code: the ISO currency code, e.g. ``"EUR"``
    :return: the amount in small currency, e.g. ``230``
    """
    validate_currency(iso_code)
    exponent = iso4217.Currency(iso_code).exponent
    if exponent == 0:
        return large_currency_amount
    return large_currency_amount * (10 ** exponent)


def to_large_currency(small_currency_amount, iso_code):
    """Inverse of :py:func:`to_small_currency`"""
    validate_currency(iso_code)
    exponent = iso4217.Currency(iso_code).exponent
    if exponent == 0:
        return small_currency_amount
    return small_currency_amount / (10 ** exponent)
